# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Path utilities for torax tests.

Paths to torax tests, etc.
"""

import pathlib
from absl.testing import absltest


def test_data_dir() -> pathlib.Path:
  """Absolute path of "test_data" directory."""
  src_dir = pathlib.Path(absltest.TEST_SRCDIR.value)
  torax_dir = 'torax/'
  path = src_dir.joinpath(torax_dir, 'tests/test_data')
  assert path.is_dir(), f'Path {path} is not a directory.'
  return path
