# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Default configs for testing."""
from typing import Any


def get_default_config_dict() -> dict[str, Any]:
  return {
      'profile_conditions': {},
      'plasma_composition': {},
      'numerics': {},
      'geometry': {'geometry_type': 'circular', 'n_rho': 4},
      'sources': {},
      'solver': {},
      'transport': {},
      'pedestal': {},
  }
